@@ -71,7 +71,7 @@
 
 #![forbid(unsafe_code)]
 
-use std::{process, sync::mpsc, fmt, time::Instant};
+use std::{borrow::Cow, fmt, process::{self, ExitCode}, sync::mpsc, time::Instant};
 
 mod args;
 mod printer;
@@ -240,6 +240,16 @@ struct TestInfo {
     is_bench: bool,
 }
 
+impl TestInfo {
+    fn test_name_with_kind(&self) -> Cow<'_, str> {
+        if self.kind.is_empty() {
+            Cow::Borrowed(&self.name)
+        } else {
+            Cow::Owned(format!("[{}] {}", self.kind, self.name))
+        }
+    }
+}
+
 /// Output of a benchmark.
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
 pub struct Measurement {
@@ -324,18 +334,13 @@ pub struct Conclusion {
 }
 
 impl Conclusion {
-    /// Exits the application with an appropriate error code (0 if all tests
-    /// have passed, 101 if there have been failures).
-    pub fn exit(&self) -> ! {
-        self.exit_if_failed();
-        process::exit(0);
-    }
-
-    /// Exits the application with error code 101 if there were any failures.
-    /// Otherwise, returns normally.
-    pub fn exit_if_failed(&self) {
+    /// Returns an exit code that can be returned from `main` to signal
+    /// success/failure to the calling process.
+    pub fn exit_code(&self) -> ExitCode {
         if self.has_failed() {
-            process::exit(101)
+            ExitCode::from(101)
+        } else {
+            ExitCode::SUCCESS
         }
     }
 
@@ -344,6 +349,25 @@ impl Conclusion {
         self.num_failed > 0
     }
 
+    /// Exits the application with an appropriate error code (0 if all tests
+    /// have passed, 101 if there have been failures). This uses
+    /// [`process::exit`], meaning that destructors are not ran. Consider
+    /// using [`Self::exit_code`] instead for a proper program cleanup.
+    pub fn exit(&self) -> ! {
+        self.exit_if_failed();
+        process::exit(0);
+    }
+
+    /// Exits the application with error code 101 if there were any failures.
+    /// Otherwise, returns normally. This uses [`process::exit`], meaning that
+    /// destructors are not ran. Consider using [`Self::exit_code`] instead for
+    /// a proper program cleanup.
+    pub fn exit_if_failed(&self) {
+        if self.has_failed() {
+            process::exit(101)
+        }
+    }
+
     fn empty() -> Self {
         Self {
             num_filtered_out: 0,
@@ -364,13 +388,23 @@ impl Arguments {
     }
 
     fn is_filtered_out(&self, test: &Trial) -> bool {
-        let test_name = &test.info.name;
+        let test_name = test.name();
+        // Match against the full test name, including the kind. This upholds the invariant that if
+        // --list prints out:
+        //
+        // <some string>: test
+        //
+        // then "--exact <some string>" runs exactly that test.
+        let test_name_with_kind = test.info.test_name_with_kind();
 
         // If a filter was specified, apply this
         if let Some(filter) = &self.filter {
             match self.exact {
-                true if test_name != filter => return true,
-                false if !test_name.contains(filter) => return true,
+                // For exact matches, we want to match against either the test name (to maintain
+                // backwards compatibility with older versions of libtest-mimic), or the test kind
+                // (technically more correct with respect to matching against the output of --list.)
+                true if test_name != filter && &test_name_with_kind != filter => return true,
+                false if !test_name_with_kind.contains(filter) => return true,
                 _ => {}
             };
         }
@@ -378,8 +412,13 @@ impl Arguments {
         // If any skip pattern were specified, test for all patterns.
         for skip_filter in &self.skip {
             match self.exact {
-                true if test_name == skip_filter => return true,
-                false if test_name.contains(skip_filter) => return true,
+                // For exact matches, we want to match against either the test name (to maintain
+                // backwards compatibility with older versions of libtest-mimic), or the test kind
+                // (technically more correct with respect to matching against the output of --list.)
+                true if test_name == skip_filter || &test_name_with_kind == skip_filter => {
+                    return true
+                }
+                false if test_name_with_kind.contains(skip_filter) => return true,
                 _ => {}
             }
         }
@@ -426,7 +465,7 @@ pub fn run(args: &Arguments, mut tests: Vec<Trial>) -> Conclusion {
 
     let mut failed_tests = Vec::new();
     let mut handle_outcome = |outcome: Outcome, test: TestInfo, printer: &mut Printer| {
-        printer.print_single_outcome(&outcome);
+        printer.print_single_outcome(&test, &outcome);
 
         // Handle outcome
         match outcome {
@@ -442,7 +481,7 @@ pub fn run(args: &Arguments, mut tests: Vec<Trial>) -> Conclusion {
 
     // Execute all tests.
     let test_mode = !args.bench;
-    if args.test_threads == Some(1) {
+    if platform_defaults_to_one_thread() || args.test_threads == Some(1) {
         // Run test sequentially in main thread
         for test in tests {
             // Print `test foo    ...`, run the test, then print the outcome in
@@ -498,6 +537,13 @@ pub fn run(args: &Arguments, mut tests: Vec<Trial>) -> Conclusion {
     conclusion
 }
 
+/// Returns whether the current host platform should use a single thread by
+/// default rather than a thread pool by default. Some platforms, such as
+/// WebAssembly, don't have native support for threading at this time.
+fn platform_defaults_to_one_thread() -> bool {
+    cfg!(target_family = "wasm")
+}
+
 /// Runs the given runner, catching any panics and treating them as a failed test.
 fn run_single(runner: Box<dyn FnOnce(bool) -> Outcome + Send>, test_mode: bool) -> Outcome {
     use std::panic::{catch_unwind, AssertUnwindSafe};
